import asyncio

import logging

from asynctnt.connection import ConnectionState
from tests import BaseTarantoolTestCase

import asynctnt


class ConnectTestCase(BaseTarantoolTestCase):
    DO_CONNECT = False

    async def test__connect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0,
                                   loop=self.loop)
        await conn.connect()
        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn._protocol.is_fully_connected())
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNotNone(conn._protocol.schema)
        self.assertIsNotNone(conn.version)
        await conn.disconnect()

    async def test__connect_no_schema(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0,
                                   fetch_schema=False,
                                   auto_refetch_schema=False,
                                   loop=self.loop)
        await conn.connect()
        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn._protocol.is_fully_connected())
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNone(conn._protocol.schema)
        await conn.disconnect()

    async def test__connect_auth(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   reconnect_timeout=0,
                                   loop=self.loop)
        await conn.connect()
        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn._protocol.is_fully_connected())
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNotNone(conn._protocol.schema)
        await conn.disconnect()

    async def test__connect_auth_no_schema(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   fetch_schema=False,
                                   auto_refetch_schema=False,
                                   reconnect_timeout=0,
                                   loop=self.loop)
        await conn.connect()
        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn._protocol.is_fully_connected())
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNone(conn._protocol.schema)
        await conn.disconnect()

    async def test__disconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0,
                                   loop=self.loop)
        await conn.connect()
        await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)
        self.assertIsNone(conn.schema)

    async def test__disconnect_auth(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   reconnect_timeout=0,
                                   loop=self.loop)
        await conn.connect()
        await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)
        self.assertIsNone(conn.schema)

    async def test__connect_multiple(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   fetch_schema=False,
                                   reconnect_timeout=0,
                                   loop=self.loop)
        for _ in range(10):
            await conn.connect()
            await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_cancel(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   fetch_schema=True,
                                   reconnect_timeout=0,
                                   loop=self.loop)
        f = asyncio.ensure_future(conn.connect(), loop=self.loop)
        await self.sleep(0.0001)
        f.cancel()
        with self.assertRaises(asyncio.CancelledError):
            await f

        await conn.disconnect()

    async def test__connect_wait_tnt_started(self):
        await self.tnt.stop()
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   fetch_schema=True,
                                   reconnect_timeout=0.000001,
                                   loop=self.loop)
        coro = self.ensure_future(conn.connect())
        await self.sleep(0.3)
        await self.tnt.start()
        await self.sleep(1)
        await coro
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        await conn.disconnect()

    async def test__connect_tnt_restarted(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   fetch_schema=True,
                                   reconnect_timeout=0.000001,
                                   loop=self.loop)
        await conn.connect()

        await self.tnt.stop()
        await self.tnt.start()
        await self.sleep(0.5)
        try:
            await conn.ping()
        except Exception as e:
            self.fail(
                'Should not throw any exceptions, but got: {}'.format(e))
        finally:
            await conn.disconnect()




import asyncio

from asynctnt import Iterator
from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class DeleteTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, 'a', 1],
            [1, 'b', 0],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__delete_one(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_ID, [data[0][0]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [data[0]], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertListEqual(res.body, [], 'Body ok')

    async def test__delete_by_name(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[0][0]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [data[0]], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertListEqual(res.body, [], 'Body ok')

    async def test__delete_by_index_id(self):
        index_name = 'temp_idx'
        res = await self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()

            data = await self._fill_data()

            res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[1][2]],
                                         index=index_id)
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertListEqual(res.body, [data[1]], 'Body ok')

            res = await self.conn.select(self.TESTER_SPACE_ID, [data[1][2]],
                                         index=index_id)
            self.assertListEqual(res.body, [], 'Body ok')
        finally:
            await self.tnt.command(
                'box.space.{}.index.{}:drop()'.format(
                    self.TESTER_SPACE_NAME, index_name)
            )

    async def test__delete_by_index_name(self):
        index_name = 'temp_idx'
        res = await self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()

            data = await self._fill_data()

            res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[1][2]],
                                         index=index_name)
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertListEqual(res.body, [data[1]], 'Body ok')

            res = await self.conn.select(self.TESTER_SPACE_ID, [data[1][2]],
                                         index=index_id)
            self.assertListEqual(res.body, [], 'Body ok')
        finally:
            await self.tnt.command(
                'box.space.{}.index.{}:drop()'.format(
                    self.TESTER_SPACE_NAME, index_name)
            )

    async def test__delete_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.delete(self.TESTER_SPACE_NAME, [0])

    async def test__delete_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.delete(self.TESTER_SPACE_ID, [0],
                                   index='primary')

    async def test__delete_invalid_types(self):
        with self.assertRaisesRegex(
                TypeError, "missing 2 required positional arguments: 'space' and 'key'"):
            await self.conn.delete()

        with self.assertRaisesRegex(
                TypeError, r'Expected list, got '):
            await self.conn.delete(self.TESTER_SPACE_ID, (1,))

        with self.assertRaisesRegex(
                TypeError, r'Expected list, got '):
            await self.conn.delete(self.TESTER_SPACE_ID, {})




include "const.pxi"

include "cmsgpuck.pxd"
include "python.pxd"

include "unicode.pxd"
include "buffer.pxd"
include "rbuffer.pxd"
include "request.pxd"
include "response.pxd"
include "schema.pxd"
include "db.pxd"

include "coreproto.pxd"


cdef class BaseProtocol(CoreProtocol):
    cdef:
        object loop
        str username
        str password
        bint fetch_schema
        bint auto_refetch_schema
        object connected_fut
        object on_connected_lost_cb

        object _on_request_completed_cb
        object _on_request_timeout_cb

        uint64_t _sync
        Schema _schema
        int64_t _schema_id
        Db _db

    cdef void _set_connection_ready(self)
    cdef void _set_connection_error(self, e)

    cdef void _do_auth(self, str username, str password)
    cdef object _do_fetch_schema(self)

    cdef uint64_t next_sync(self)
    cdef uint32_t transform_iterator(self, iterator) except *
    cdef uint32_t transform_space(self, space) except *
    cdef uint32_t transform_index(self, space, index) except *

    cdef object _new_waiter_for_request(self, Request req, float timeout)
    cdef Db _create_db(self)
    cdef object execute(self, Request req, float timeout)



import enum


class TarantoolError(Exception):
    pass


class TarantoolSchemaError(TarantoolError):
    pass


class TarantoolDatabaseError(TarantoolError):

    def __init__(self, code, message):
        super(TarantoolDatabaseError, self).__init__(code, message)
        self.code = code
        self.error_code = ErrorCode(code)
        self.message = message


class TarantoolNetworkError(TarantoolError):
    pass


class TarantoolNotConnectedError(TarantoolNetworkError):
    pass


class TarantoolConnectionLostError(TarantoolNetworkError):
    pass


class ErrorCode(enum.IntEnum):
    ER_UNKNOWN = 0
    ER_ILLEGAL_PARAMS = 1
    ER_MEMORY_ISSUE = 2
    ER_TUPLE_FOUND = 3
    ER_TUPLE_NOT_FOUND = 4
    ER_UNSUPPORTED = 5
    ER_NONMASTER = 6
    ER_READONLY = 7
    ER_INJECTION = 8
    ER_CREATE_SPACE = 9
    ER_SPACE_EXISTS = 10
    ER_DROP_SPACE = 11
    ER_ALTER_SPACE = 12
    ER_INDEX_TYPE = 13
    ER_MODIFY_INDEX = 14
    ER_LAST_DROP = 15
    ER_TUPLE_FORMAT_LIMIT = 16
    ER_DROP_PRIMARY_KEY = 17
    ER_KEY_PART_TYPE = 18
    ER_EXACT_MATCH = 19
    ER_INVALID_MSGPACK = 20
    ER_PROC_RET = 21
    ER_TUPLE_NOT_ARRAY = 22
    ER_FIELD_TYPE = 23
    ER_FIELD_TYPE_MISMATCH = 24
    ER_SPLICE = 25
    ER_ARG_TYPE = 26
    ER_TUPLE_IS_TOO_LONG = 27
    ER_UNKNOWN_UPDATE_OP = 28
    ER_UPDATE_FIELD = 29
    ER_FIBER_STACK = 30
    ER_KEY_PART_COUNT = 31
    ER_PROC_LUA = 32
    ER_NO_SUCH_PROC = 33
    ER_NO_SUCH_TRIGGER = 34
    ER_NO_SUCH_INDEX = 35
    ER_NO_SUCH_SPACE = 36
    ER_NO_SUCH_FIELD = 37
    ER_EXACT_FIELD_COUNT = 38
    ER_INDEX_FIELD_COUNT = 39
    ER_WAL_IO = 40
    ER_MORE_THAN_ONE_TUPLE = 41
    ER_ACCESS_DENIED = 42
    ER_CREATE_USER = 43
    ER_DROP_USER = 44
    ER_NO_SUCH_USER = 45
    ER_USER_EXISTS = 46
    ER_PASSWORD_MISMATCH = 47
    ER_UNKNOWN_REQUEST_TYPE = 48
    ER_UNKNOWN_SCHEMA_OBJECT = 49
    ER_CREATE_FUNCTION = 50
    ER_NO_SUCH_FUNCTION = 51
    ER_FUNCTION_EXISTS = 52
    ER_FUNCTION_ACCESS_DENIED = 53
    ER_FUNCTION_MAX = 54
    ER_SPACE_ACCESS_DENIED = 55
    ER_USER_MAX = 56
    ER_NO_SUCH_ENGINE = 57
    ER_RELOAD_CFG = 58
    ER_CFG = 59
    ER_VINYL = 60
    ER_LOCAL_SERVER_IS_NOT_ACTIVE = 61
    ER_UNKNOWN_SERVER = 62
    ER_CLUSTER_ID_MISMATCH = 63
    ER_INVALID_UUID = 64
    ER_CLUSTER_ID_IS_RO = 65
    ER_SERVER_ID_MISMATCH = 66
    ER_SERVER_ID_IS_RESERVED = 67
    ER_INVALID_ORDER = 68
    ER_MISSING_REQUEST_FIELD = 69
    ER_IDENTIFIER = 70
    ER_DROP_FUNCTION = 71
    ER_ITERATOR_TYPE = 72
    ER_REPLICA_MAX = 73
    ER_INVALID_XLOG = 74
    ER_INVALID_XLOG_NAME = 75
    ER_INVALID_XLOG_ORDER = 76
    ER_NO_CONNECTION = 77
    ER_TIMEOUT = 78
    ER_ACTIVE_TRANSACTION = 79
    ER_NO_ACTIVE_TRANSACTION = 80
    ER_CROSS_ENGINE_TRANSACTION = 81
    ER_NO_SUCH_ROLE = 82
    ER_ROLE_EXISTS = 83
    ER_CREATE_ROLE = 84
    ER_INDEX_EXISTS = 85
    ER_TUPLE_REF_OVERFLOW = 86
    ER_ROLE_LOOP = 87
    ER_GRANT = 88
    ER_PRIV_GRANTED = 89
    ER_ROLE_GRANTED = 90
    ER_PRIV_NOT_GRANTED = 91
    ER_ROLE_NOT_GRANTED = 92
    ER_MISSING_SNAPSHOT = 93
    ER_CANT_UPDATE_PRIMARY_KEY = 94
    ER_UPDATE_INTEGER_OVERFLOW = 95
    ER_GUEST_USER_PASSWORD = 96
    ER_TRANSACTION_CONFLICT = 97
    ER_UNSUPPORTED_ROLE_PRIV = 98
    ER_LOAD_FUNCTION = 99
    ER_FUNCTION_LANGUAGE = 100
    ER_RTREE_RECT = 101
    ER_PROC_C = 102
    ER_UNKNOWN_RTREE_INDEX_DISTANCE_TYPE = 103
    ER_PROTOCOL = 104
    ER_UPSERT_UNIQUE_SECONDARY_KEY = 105
    ER_WRONG_INDEX_RECORD = 106
    ER_WRONG_INDEX_PARTS = 107
    ER_WRONG_INDEX_OPTIONS = 108
    ER_WRONG_SCHEMA_VERSION = 109
    ER_SLAB_ALLOC_MAX = 110
    ER_WRONG_SPACE_OPTIONS = 111
    ER_UNSUPPORTED_INDEX_FEATURE = 112
    ER_VIEW_IS_RO = 113
    ER_SERVER_UUID_MISMATCH = 114
    ER_SYSTEM = 115
    ER_LOADING = 116
    ER_CONNECTION_TO_SELF = 117
    ER_KEY_PART_IS_TOO_LONG = 118
    ER_COMPRESSION = 119



from libc.stdint cimport int64_t


cdef class SchemaIndex:
    cdef public int sid
    cdef public int iid
    cdef public str name
    cdef public str index_type
    cdef public object unique
    cdef public list parts


cdef class SchemaSpace:
    cdef public int sid
    cdef public int arity
    cdef public str name
    cdef public dict indexes

    cdef add_index(self, SchemaIndex idx)


cdef class Schema:
    cdef:
        dict schema
        int64_t id

    cpdef get_id(self)
    cpdef SchemaSpace get_space(self, space)
    cpdef SchemaIndex get_index(self, space, index)

    cdef inline clear(self)


cdef Schema parse_schema(int64_t schema_id, spaces, indexes)



from libc.stdint cimport uint32_t, uint64_t, int64_t

cimport tnt


cdef class WriteBuffer:
    cdef:
        # Preallocated small buffer
        bint _smallbuf_inuse
        char _smallbuf[_BUFFER_INITIAL_SIZE]

        char *_buf
        ssize_t _size  # Allocated size
        ssize_t _length  # Length of data in the buffer
        int _view_count  # Number of memoryviews attached to the buffer

        bytes _encoding

    @staticmethod
    cdef WriteBuffer new(bytes encoding=*)

    cdef inline _check_readonly(self)
    cdef inline len(self)
    cdef void ensure_allocated(self, ssize_t extra_length) except *
    cdef char *_ensure_allocated(self, char *p,
                                        ssize_t extra_length) except NULL
    cdef void _reallocate(self, ssize_t new_size) except *
    cdef void write_buffer(self, WriteBuffer buf) except *
    cdef void write_header(self, uint64_t sync, tnt.tp_request_type op,
                           int64_t schema_id=*) except *
    cdef void write_length(self)

    cdef char *_encode_nil(self, char *p) except NULL
    cdef char *_encode_bool(self, char *p, bint value) except NULL
    cdef char *_encode_double(self, char *p, double value) except NULL
    cdef char *_encode_uint(self, char *p, uint64_t value) except NULL
    cdef char *_encode_int(self, char *p, int64_t value) except NULL
    cdef char *_encode_str(self, char *p,
                           const char *str, uint32_t len) except NULL
    cdef char *_encode_bin(self, char *p,
                           const char *data, uint32_t len) except NULL
    cdef char *_encode_array(self, char *p, uint32_t len) except NULL
    cdef char *_encode_map(self, char *p, uint32_t len) except NULL
    cdef char *_encode_list(self, char *p, list arr) except NULL
    cdef char *_encode_tuple(self, char *p, tuple t) except NULL
    cdef char *_encode_dict(self, char *p, dict d) except NULL
    cdef char *_encode_obj(self, char *p, object o) except NULL

    cdef tnt.tnt_update_op_kind _op_type_to_kind(self, char *str, ssize_t len)
    cdef char *_encode_update_ops(self, char *p, list operations) except NULL

    cdef void encode_request_call(self, str func_name, list args) except *
    cdef void encode_request_eval(self, str expression, list args) except *
    cdef void encode_request_select(self, uint32_t space, uint32_t index,
                                    list key, uint64_t offset, uint64_t limit,
                                    uint32_t iterator) except *
    cdef void encode_request_insert(self, uint32_t space, list t) except *
    cdef void encode_request_delete(self, uint32_t space, uint32_t index,
                                    list key) except *
    cdef void encode_request_update(self, uint32_t space, uint32_t index,
                                    list key_tuple, list operations,
                                    uint32_t key_of_tuple=*,
                                    uint32_t key_of_operations=*
                                    ) except *
    cdef void encode_request_upsert(self, uint32_t space,
                                    list t, list operations) except *
    cdef void encode_request_auth(self,
                                  bytes username,
                                  bytes scramble) except *



from libc.stdint cimport uint64_t, uint32_t, int64_t


cdef class Response:
    cdef:
        public uint32_t code
        public uint64_t sync
        public int64_t schema_id
        public str errmsg
        public list body
        public bytes encoding

    @staticmethod
    cdef inline Response new(bytes encoding)

    cdef inline has_schema_id(self)
    cdef inline is_error(self)


cdef Response response_parse(const char *buf, uint32_t buf_len, bytes encoding)



cimport cython
cimport cpython.dict

import asyncio
import enum

from asynctnt.exceptions import \
    TarantoolSchemaError, TarantoolNotConnectedError

include "const.pxi"

include "unicode.pyx"
include "buffer.pyx"
include "rbuffer.pyx"
include "request.pyx"
include "response.pyx"
include "schema.pyx"
include "db.pyx"

include "coreproto.pyx"


class Iterator(enum.IntEnum):
    EQ = 0
    REQ = 1
    ALL = 2
    LT = 3
    LE = 4
    GE = 5
    GT = 6
    BITS_ALL_SET = 7
    BITS_ANY_SET = 8
    BITS_ALL_NOT_SET = 9
    OVERLAPS = 10
    NEIGHBOR = 11


cdef class BaseProtocol(CoreProtocol):
    def __init__(self, host, port,
                 username, password,
                 fetch_schema,
                 auto_refetch_schema,
                 connected_fut,
                 on_connection_made, on_connection_lost,
                 loop,
                 request_timeout=None,
                 encoding=None,
                 initial_read_buffer_size=None):
        CoreProtocol.__init__(self, host, port, encoding,
                              initial_read_buffer_size)

        self.loop = loop

        self.username = username
        self.password = password
        self.fetch_schema = fetch_schema
        self.auto_refetch_schema = auto_refetch_schema
        self.request_timeout = request_timeout or 0
        self.connected_fut = connected_fut
        self.on_connection_made_cb = on_connection_made
        self.on_connection_lost_cb = on_connection_lost

        self._on_request_completed_cb = self._on_request_completed
        self._on_request_timeout_cb = self._on_request_timeout

        self._sync = 0
        self._schema = None
        self._schema_id = -1
        self._db = self._create_db()

        try:
            self.create_future = self.loop.create_future
        except AttributeError:
            self.create_future = self._create_future_fallback

    def _create_future_fallback(self):  # pragma: no cover
        return asyncio.Future(loop=self.loop)

    @property
    def schema(self):
        return self._schema

    @property
    def schema_id(self):
        return self._schema_id

    cdef void _set_connection_ready(self):
        if not self.connected_fut.done():
            self.connected_fut.set_result(True)
            self.con_state = CONNECTION_FULL

    cdef void _set_connection_error(self, e):
        if not self.connected_fut.done():
            self.connected_fut.set_exception(e)
            self.con_state = CONNECTION_BAD

    cdef void _on_greeting_received(self):
        if self.username and self.password:
            self._do_auth(self.username, self.password)
        elif self.fetch_schema:
            self._do_fetch_schema()
        else:
            self._set_connection_ready()

    cdef void _do_auth(self, str username, str password):
        # No extra error handling from Db.execute
        fut = self.execute(
            self._db._auth(self.salt, username, password),
            0
        )

        def on_authorized(f):
            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                logger.debug('Tarantool[%s:%s] Authorized successfully',
                             self.host, self.port)

                if self.fetch_schema:
                    self._do_fetch_schema()
                else:
                    self._set_connection_ready()
            else:
                logger.error('Tarantool[%s:%s] Authorization failed: %s',
                             self.host, self.port, str(e))
                self._set_connection_error(e)

        fut.add_done_callback(on_authorized)

    cdef object _do_fetch_schema(self):
        fut = self.create_future()

        def on_fetch(f):
            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                spaces, indexes = f.result()
                logger.debug('Tarantool[%s:%s] Schema fetch succeeded. '
                             'Spaces: %d, Indexes: %d.',
                             self.host, self.port,
                             len(spaces.body), len(indexes.body))
                self._schema = parse_schema(spaces.schema_id,
                                            spaces.body, indexes.body)
                if self.auto_refetch_schema:
                    # if no refetch, them we should not
                    # send schema_id at all (leave it -1)
                    self._schema_id = self._schema.id
                else:
                    self._schema_id = -1
                self._set_connection_ready()
                fut.set_result(self._schema)
            else:
                logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
                             self.host, self.port, str(e))
                if isinstance(e, asyncio.TimeoutError):
                    e = asyncio.TimeoutError('Schema fetch timeout')
                self._set_connection_error(e)
                fut.set_exception(e)

        self._schema_id = -1
        fut_vspace = self._db.select(SPACE_VSPACE, timeout=0)
        fut_vindex = self._db.select(SPACE_VINDEX, timeout=0)
        gather_fut = asyncio.gather(fut_vspace, fut_vindex,
                                    return_exceptions=False,
                                    loop=self.loop)
        gather_fut.add_done_callback(on_fetch)
        return fut

    cdef void _on_connection_made(self):
        CoreProtocol._on_connection_made(self)

        if self.on_connection_made_cb:
            self.on_connection_made_cb()

    cdef void _on_connection_lost(self, exc):
        CoreProtocol._on_connection_lost(self, exc)

        if self.on_connection_lost_cb:
            self.on_connection_lost_cb(exc)

    cdef uint64_t next_sync(self):
        self._sync += 1
        return self._sync

    def _on_request_timeout(self, waiter):
        cdef Request req

        if waiter.done():
            return

        req = waiter._req
        req.timeout_handle.cancel()
        req.timeout_handle = None
        waiter.set_exception(
            asyncio.TimeoutError(
                '{} exceeded timeout'.format(req.__class__.__name__))
        )

    def _on_request_completed(self, fut):
        cdef Request req = fut._req
        fut._req = None

        if req.timeout_handle is not None:
            req.timeout_handle.cancel()
            req.timeout_handle = None

    cdef object _new_waiter_for_request(self, Request req, float timeout):
        fut = self.create_future()
        fut._req = req  # to be able to retrieve request after done()
        req.waiter = fut

        if timeout < 0:
            timeout = self.request_timeout
        if timeout is not None and timeout > 0:
            req.timeout_handle = \
                self.loop.call_later(timeout,
                                     self._on_request_timeout_cb, fut)
            fut.add_done_callback(self._on_request_completed_cb)
        return fut

    cdef Db _create_db(self):
        return Db.new(self)

    def create_db(self):
        return self._create_db()

    def get_common_db(self):
        return self._db

    cdef object execute(self, Request req, float timeout):
        if self.con_state == CONNECTION_BAD:
            raise TarantoolNotConnectedError('Tarantool is not connected')

        cpython.dict.PyDict_SetItem(self.reqs, req.sync, req)
        self._write(req.build())

        return self._new_waiter_for_request(req, timeout)

    cdef uint32_t transform_iterator(self, iterator) except *:
        if isinstance(iterator, int):
            return iterator
        if isinstance(iterator, Iterator):
            return iterator.value
        if isinstance(iterator, str):
            return Iterator[iterator]
        else:
            raise TypeError('Iterator is of unsupported type '
                            '(asynctnt.Iterator, int, str)')

    cdef uint32_t transform_space(self, space) except *:
        if isinstance(space, str):
            if self._schema is None:
                raise TarantoolSchemaError('Schema not fetched')

            sp = self._schema.get_space(space)
            if sp is None:
                raise TarantoolSchemaError('Space {} not found'.format(space))
            return sp.sid
        return space

    cdef uint32_t transform_index(self, space, index) except *:
        if isinstance(index, str):
            if self._schema is None:
                raise TarantoolSchemaError('Schema not fetched')

            idx = self._schema.get_index(space, index)
            if idx is None:
                raise TarantoolSchemaError(
                    'Index {} for space {} not found'.format(index, space))
            return idx.iid
        return index

    def refetch_schema(self):
        return self._do_fetch_schema()


class Protocol(BaseProtocol, asyncio.Protocol):
    pass



cdef class SchemaIndex:

    def __init__(self, list index_row):
        self.sid = index_row[0]
        self.iid = index_row[1]
        self.name = index_row[2]
        self.index_type = index_row[3]
        self.unique = index_row[4]
        self.parts = []
        if isinstance(index_row[5], (list, tuple)):
            for k, v in index_row[5]:
                self.parts.append((k, v))
        else:
            for i in range(index_row[5]):
                self.parts.append(
                    (index_row[5 + 1 + i * 2], index_row[5 + 2 + i * 2]))

        def __repr__(self):
            return \
                '<SchemaIndex sid={}, id={}, name={}, ' \
                'type={}, unique={}>'.format(
                    self.sid, self.iid, self.name, self.index_type, self.unique
                )


cdef class SchemaSpace:
    def __init__(self, list space_row):
        self.sid = space_row[0]
        self.arity = space_row[1]
        self.name = space_row[2]
        self.indexes = {}

    cdef add_index(self, SchemaIndex idx):
        self.indexes[idx.iid] = idx
        if idx.name:
            self.indexes[idx.name] = idx

    def __repr__(self):
        return '<SchemaSpace id={}, name={}, arity={}>'.format(
            self.sid, self.name, self.arity
        )


cdef class Schema:
    def __init__(self, int64_t schema_id):
        self.id = schema_id
        self.schema = {}

    cpdef SchemaSpace get_space(self, space):
        try:
            return self.schema[space]
        except KeyError:
            return None

    cpdef SchemaIndex get_index(self, space, index):
        sp = self.get_space(space)
        if sp is None:
            return None
        try:
            return sp.indexes[index]
        except KeyError:
            return None

    cpdef get_id(self):
        return self.id

    cdef inline clear(self):
        self.schema.clear()

    def __repr__(self):
        return '<Schema>'


cdef Schema parse_schema(int64_t schema_id, spaces, indexes):
    cdef:
        Schema s
        SchemaSpace sp
        SchemaIndex idx
    s = Schema(schema_id)
    for space_row in spaces:
        sp = SchemaSpace(space_row)
        s.schema[sp.sid] = sp
        if sp.name:
            s.schema[sp.name] = sp

    for index_row in indexes:
        idx = SchemaIndex(index_row)
        sp = s.schema[idx.sid]
        sp.add_index(idx)
    return s



import logging

import asynctnt
from asynctnt.exceptions import TarantoolNotConnectedError

from tests import BaseTarantoolTestCase
from tests.util import get_complex_param, get_big_param


class CommonTestCase(BaseTarantoolTestCase):
    async def test__encoding_utf8(self):
        p, p_cmp = get_complex_param(replace_bin=False)

        data = [1, 'hello', p]
        data_cmp = [1, 'hello', p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertListEqual(res.body, [data_cmp], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertListEqual(res.body, [data_cmp], 'Body ok')

    async def test__encoding_cp1251(self):
        await self.tnt_reconnect(encoding='cp1251')
        p, p_cmp = get_complex_param(replace_bin=False)

        data = [1, 'hello', p]
        data_cmp = [1, 'hello', p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertListEqual(res.body, [data_cmp], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertListEqual(res.body, [data_cmp], 'Body ok')

    async def test__schema_refetch_on_schema_change(self):
        await self.tnt_reconnect(auto_refetch_schema=True)
        self.assertTrue(self.conn.fetch_schema)
        self.assertTrue(self.conn.auto_refetch_schema)
        schema_before = self.conn.schema_id
        self.assertNotEqual(schema_before, -1)

        # Changing scheme
        await self.conn.eval(
            "s = box.schema.create_space('new_space');"
            "s:drop();"
        )

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

        self.assertGreater(self.conn.schema_id, schema_before,
                           'Schema changed')

    async def test__schema_no_fetch_and_refetch(self):
        await self.tnt_reconnect(auto_refetch_schema=False,
                                 fetch_schema=False)
        self.assertFalse(self.conn.fetch_schema)
        self.assertFalse(self.conn.auto_refetch_schema)
        self.assertEqual(self.conn.schema_id, -1)

        # Changing scheme
        await self.conn.eval(
            "s = box.schema.create_space('new_space');"
            "s:drop();"
        )

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

    async def test__read_buffer_reallocate_ok(self):
        await self.tnt_reconnect(initial_read_buffer_size=1)

        p, cmp = get_complex_param(encoding=self.conn.encoding)
        try:
            res = await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

        self.assertDictEqual(res.body[0][0], cmp, 'Body ok')

    async def test__read_buffer_deallocate_ok(self):
        size = 100 * 1000
        await self.tnt_reconnect(initial_read_buffer_size=size)

        # Waiting big response, so ReadBuffer grows to hold it
        p = get_big_param(size=size * 3)
        try:
            res = await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

        # Waiting small response, so ReadBuffer deallocates memory
        p = get_big_param(size=10)
        try:
            res = await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

    async def test__write_buffer_reallocate(self):
        p = get_big_param(size=100*1024)
        try:
            res = await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

        self.assertDictEqual(res.body[0][0], p, 'Body ok')

    async def test__ensure_no_attribute_error_on_not_connected(self):
        await self.tnt_disconnect()

        self.conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            loop=self.loop)

        with self.assertRaises(TarantoolNotConnectedError):
            await self.conn.ping()

    async def test__encode_unsupported_type(self):
        class A:
            pass
        with self.assertRaisesRegex(TypeError,
                                    'Type `(.+)` is not supported for encoding'):
            self.conn.call('func_param', [{'a': A()}])



import asyncio

from asynctnt import Iterator
from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError, TarantoolDatabaseError, \
    ErrorCode
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class UpdateTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, 'a', 1, 5],
            [1, 'b', 0, 6],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__update_one_assign(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['=', 2, 2]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')

        data[1][2] = 2
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_insert(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['!', 2, 14]])
        data[1].insert(2, 14)
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_delete(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['#', 2, 1]])
        data[1].pop(2)
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_plus(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['+', 2, 3]])
        data[1][2] += 3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_plus_negative(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['+', 2, -3]])
        data[1][2] += -3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_minus(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['-', 2, 3]])
        data[1][2] -= 3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_minus_negative(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['-', 2, -3]])
        data[1][2] -= -3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_band(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['&', 2, 3]])
        data[1][2] &= 3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['&', 2, 2]])
        data[1][2] &= 2
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['&', 2, 1]])
        data[1][2] &= 1
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['&', 2, 0]])
        data[1][2] &= 0
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_bor(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['|', 2, 3]])
        data[1][2] |= 3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['|', 2, 2]])
        data[1][2] |= 2
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['|', 2, 1]])
        data[1][2] |= 1
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['|', 2, 0]])
        data[1][2] |= 0
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_one_bxor(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['^', 2, 3]])
        data[1][2] ^= 3
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['^', 2, 2]])
        data[1][2] ^= 2
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['^', 2, 1]])
        data[1][2] ^= 1
        self.assertListEqual(res.body, [data[1]], 'Body ok')

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [['^', 2, 0]])
        data[1][2] ^= 0
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_splice(self):
        data = [1, 'hello', 'hi']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [[':', 2, 1, 3, '!!!']])

        data[2] = 'h!!!'
        self.assertListEqual(res.body, [data], 'Body ok')

    async def test__update_splice_bytes(self):
        data = [1, 'hello', 'hi']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], [[b':', 2, 1, 3, '!!!']])

        data[2] = 'h!!!'
        self.assertListEqual(res.body, [data], 'Body ok')

    async def test__update_splice_wrong_args(self):
        data = [1, 'hello', 'hi']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        with self.assertRaisesRegex(
                IndexError, r'Operation length must be at least 3'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[':', 2]])

        with self.assertRaisesRegex(
                IndexError, r'Splice operation must have length of 5'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[':', 2, 1]])

        with self.assertRaisesRegex(
                IndexError, r'Splice operation must have length of 5'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[':', 2, 1, 3]])

        with self.assertRaisesRegex(
                TypeError, r'Splice offset must be int'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[':', 2, 1, {}, ':::']])

        with self.assertRaisesRegex(
                TypeError, r'Splice position must be int'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[':', 2, {}, {}, ':::']])

        with self.assertRaisesRegex(
                TypeError, r'Operation field_no must be of int type'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[':', {}, {}, {}, ':::']])

        with self.assertRaisesRegex(
                TypeError, r'Unknown update operation type `yo`'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [['yo', 1, 2, 3]])

        with self.assertRaisesRegex(
                TypeError, r'Operation type must of a str or bytes type'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [[{}, 1, 2, 3]])

        with self.assertRaisesRegex(
                TypeError, r'Single operation must be a tuple or list'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [{}])

        with self.assertRaisesRegex(
                TypeError, r'int argument required for '
                           r'Arithmetic and Delete operations'):
            await self.conn.update(self.TESTER_SPACE_ID,
                                   [1], [('+', 2, {})])

    async def test__update_multiple_operations(self):
        t = [1, '1', 1, 0, 3, 4, 8, 'hello']
        await self.conn.insert(self.TESTER_SPACE_ID, t)

        t[2] += 1
        t[3] -= 4
        t[4] &= 5
        t[5] |= 7
        t[6] = 100
        t[7] = 'h!!!o'

        operations = [
            ['+', 2, 1],
            ['-', 3, 4],
            ['&', 4, 5],
            ['|', 5, 7],
            ['=', 6, 100],
            [':', 7, 1, 3, '!!!'],
        ]

        res = await self.conn.update(self.TESTER_SPACE_ID,
                                     [1], operations)
        self.assertListEqual(res.body, [t], 'Body ok')

    async def test__update_by_name(self):
        data = await self._fill_data()

        res = await self.conn.update(self.TESTER_SPACE_NAME,
                                     [1], [['=', 2, 2]])

        data[1][2] = 2
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [data[1]], 'Body ok')

    async def test__update_by_name_no_schema(self):
        await self._fill_data()

        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.update(self.TESTER_SPACE_NAME,
                                   [1], [['=', 2, 2]])

    async def test__update_by_index_id(self):
        index_name = 'temp_idx'
        res = await self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()
            data = await self._fill_data()

            res = await self.conn.update(self.TESTER_SPACE_ID, [data[0][2]],
                                         [('=', 2, 1)], index=index_id)

            data[0][2] = 1
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertListEqual(res.body, [data[0]], 'Body ok')
        finally:
            await self.tnt.command(
                'box.space.{}.index.{}:drop()'.format(
                    self.TESTER_SPACE_NAME, index_name)
            )

    async def test__select_by_index_name(self):
        index_name = 'temp_idx'
        res = await self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()
            data = await self._fill_data()

            res = await self.conn.update(self.TESTER_SPACE_ID, [data[0][2]],
                                         [['=', 2, 1]], index=index_name)

            data[0][2] = 1
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertListEqual(res.body, [data[0]], 'Body ok')
        finally:
            await self.tnt.command(
                'box.space.{}.index.{}:drop()'.format(
                    self.TESTER_SPACE_NAME, index_name)
            )

    async def test__update_by_index_id_no_schema(self):
        await self._fill_data()
        await self.tnt_reconnect(fetch_schema=False)

        try:
            await self.conn.update(self.TESTER_SPACE_ID, [0],
                                   [['=', 2, 1]], index=0)
        except Exception as e:
            self.fail(e)

    async def test__update_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.update(self.TESTER_SPACE_NAME, [0],
                                   [['=', 2, 1]], index='primary')

    async def test__update_operations_none(self):
        data = await self._fill_data()
        try:
            res = await self.conn.update(self.TESTER_SPACE_NAME,
                                         [data[0][0]], None)
        except TarantoolDatabaseError as e:
            if self.conn.version < (1, 7):
                if e.code == ErrorCode.ER_ILLEGAL_PARAMS:
                    # success
                    return
            raise
        self.assertListEqual(res.body, [data[0]], 'empty operations')



from libc.stdint cimport uint32_t, uint64_t, int64_t

cimport tnt

cdef class Request:
    cdef:
        tnt.tp_request_type op
        uint64_t sync
        int64_t schema_id
        WriteBuffer buf
        WriteBuffer buf_body
        object waiter
        object timeout_handle

    @staticmethod
    cdef inline Request new(tnt.tp_request_type op,
                            uint64_t sync, int64_t schema_id,
                            WriteBuffer buf_body)

    cdef WriteBuffer build(self)



cimport cpython
cimport tnt

import hashlib
from asynctnt.exceptions import TarantoolDatabaseError, ErrorCode


cdef class Db:
    def __cinit__(self):
        self._protocol = None
        self._encoding = None

    @staticmethod
    cdef inline Db new(BaseProtocol protocol):
        cdef Db db = Db.__new__(Db)
        db._protocol = protocol
        db._encoding = protocol.encoding
        return db

    cdef inline uint64_t next_sync(self):
        return self._protocol.next_sync()

    async def execute(self, Request req, float timeout):
        cdef object fut
        try:
            return await self._protocol.execute(req, timeout)
        except TarantoolDatabaseError as e:
            if e.code == ErrorCode.ER_WRONG_SCHEMA_VERSION:
                await self._protocol.refetch_schema()

                # Retry request with updated schema_id
                req.schema_id = self._protocol._schema_id
                return await self._protocol.execute(req, timeout)
            raise

    cdef Request _ping(self):
        return Request.new(tnt.TP_PING,
                           self.next_sync(),
                           self._protocol._schema_id,
                           None)

    cdef Request _call16(self, str func_name, list args):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_call(func_name, args)
        return Request.new(tnt.TP_CALL_16,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _call(self, str func_name, list args):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_call(func_name, args)
        return Request.new(tnt.TP_CALL,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _eval(self, str expression, list args):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_eval(expression, args)
        return Request.new(tnt.TP_EVAL,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _select(self, uint32_t space, uint32_t index, list key,
                         uint64_t offset, uint64_t limit, uint32_t iterator):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_select(space, index, key,
                                  offset, limit, iterator)
        return Request.new(tnt.TP_SELECT,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _insert(self, uint32_t space, list t, bint replace):
        cdef:
            tnt.tp_request_type op
            WriteBuffer buf

        op = tnt.TP_INSERT if not replace else tnt.TP_REPLACE
        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_insert(space, t)
        return Request.new(op,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _delete(self, uint32_t space, uint32_t index, list key):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_delete(space, index, key)
        return Request.new(tnt.TP_DELETE,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _update(self, uint32_t space, uint32_t index,
                         list key, list operations):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_update(space, index, key, operations)
        return Request.new(tnt.TP_UPDATE,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _upsert(self, uint32_t space, list t, list operations):
        cdef:
            WriteBuffer buf

        buf = WriteBuffer.new(self._encoding)
        buf.encode_request_upsert(space, t, operations)
        return Request.new(tnt.TP_UPSERT,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    cdef Request _auth(self, bytes salt, str username, str password):
        cdef:
            WriteBuffer buf

            bytes username_bytes, password_bytes
            bytes hash1, hash2, scramble

        buf = WriteBuffer.new(self._encoding)

        username_bytes = encode_unicode_string(username, self._encoding)
        password_bytes = encode_unicode_string(password, self._encoding)

        hash1 = Db._sha1((password_bytes,))
        hash2 = Db._sha1((hash1,))
        scramble = Db._sha1((salt, hash2))
        scramble = Db._strxor(hash1, scramble)

        buf.encode_request_auth(username_bytes, scramble)
        return Request.new(tnt.TP_AUTH,
                           self.next_sync(),
                           self._protocol._schema_id,
                           buf)

    @staticmethod
    cdef bytes _sha1(tuple values):
        cdef object sha = hashlib.sha1()
        for i in values:
            if i is not None:
                sha.update(i)
        return sha.digest()

    @staticmethod
    cdef bytes _strxor(bytes hash1, bytes scramble):
        cdef:
            char *hash1_str
            ssize_t hash1_len

            char *scramble_str
            ssize_t scramble_len
        cpython.bytes.PyBytes_AsStringAndSize(hash1,
                                              &hash1_str, &hash1_len)
        cpython.bytes.PyBytes_AsStringAndSize(scramble,
                                              &scramble_str, &scramble_len)
        for i in range(scramble_len):
            scramble_str[i] = hash1_str[i] ^ scramble_str[i]
        return scramble

    # public methods

    def ping(self, timeout=-1):
        return self.execute(
            self._ping(),
            timeout
        )

    def call16(self, func_name, args=None, timeout=-1):
        return self.execute(
            self._call16(func_name, args),
            timeout
        )

    def call(self, func_name, args=None, timeout=-1):
        return self.execute(
            self._call(func_name, args),
            timeout
        )

    def eval(self, expression, args=None, timeout=-1):
        return self.execute(
            self._eval(expression, args),
            timeout
        )

    def select(self, space, key=None,
                 offset=0, limit=0xffffffff, index=0, iterator=0, timeout=-1):
        space = self._protocol.transform_space(space)
        index = self._protocol.transform_index(space, index)
        iterator = self._protocol.transform_iterator(iterator)

        return self.execute(
            self._select(space, index, key, offset, limit, iterator),
            timeout
        )

    def insert(self, space, t, replace=False, timeout=-1):
        space = self._protocol.transform_space(space)

        return self.execute(
            self._insert(space, t, replace),
            timeout
        )

    def replace(self, space, t, timeout=-1):
        return self.insert(space, t, replace=True, timeout=timeout)

    def delete(self, space, key, index=0, timeout=-1):
        space = self._protocol.transform_space(space)
        index = self._protocol.transform_index(space, index)

        return self.execute(
            self._delete(space, index, key),
            timeout
        )

    def update(self, space, key, operations, index=0, timeout=-1):
        space = self._protocol.transform_space(space)
        index = self._protocol.transform_index(space, index)

        return self.execute(
            self._update(space, index, key, operations),
            timeout
        )

    def upsert(self, space, t, operations, timeout=-1):
        space = self._protocol.transform_space(space)

        return self.execute(
            self._upsert(space, t, operations),
            timeout
        )



import asyncio

import sys

import logging

from asynctnt import Response
from asynctnt.exceptions import TarantoolNotConnectedError
from tests import BaseTarantoolTestCase


class PingTestCase(BaseTarantoolTestCase):
    SMALL_TIMEOUT = 0.00000000001
    LOGGING_LEVEL = logging.DEBUG

    async def test__ping_basic(self):
        res = await self.conn.ping()
        self.assertIsNotNone(res)
        self.assertIsInstance(res, Response)
        self.assertGreater(res.sync, 0, 'Sync is not 0')
        self.assertEqual(res.code, 0, 'Code is 0')
        self.assertIsNone(res.body, 'No body for ping')

    async def test__ping_timeout_on_conn(self):
        await self.tnt_reconnect(request_timeout=self.SMALL_TIMEOUT)

        try:
            await self.conn.ping(timeout=1)
        except:
            self.fail('Should not fail on timeout 1')

    async def test__ping_connection_lost(self):
        await self.tnt.stop()

        with self.assertRaises(TarantoolNotConnectedError):
            await self.conn.ping()

        await self.tnt.start()
        await self.sleep(1)

        try:
            await self.conn.ping()
        except:
            self.fail('Should not fail after tnt start')

    async def test__ping_with_reconnect(self):
        await self.conn.reconnect()
        res = await self.conn.ping()
        self.assertIsInstance(res, Response, 'Ping result')



import asyncio

from asynctnt import Iterator
from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class InsertTestCase(BaseTarantoolTestCase):
    async def test__insert_one(self):
        data = [1, 'hello']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [data], 'Body ok')

    async def test__insert_by_name(self):
        data = [1, 'hello']
        res = await self.conn.insert(self.TESTER_SPACE_NAME, data)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertListEqual(res.body, [data], 'Body ok')

    async def test__insert_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        data = [1, 'hello']
        with self.assertRaises(TarantoolSchemaError):
            await self.conn.insert(self.TESTER_SPACE_NAME, data)

    async def test__insert_complex_tuple(self):
        p, p_cmp = get_complex_param(replace_bin=False)
        data = [1, 'hello', p]
        data_cmp = [1, 'hello', p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertListEqual(res.body, [data_cmp], 'Body ok')

    async def test__insert_replace(self):
        data = [1, 'hello']

        await self.conn.insert(self.TESTER_SPACE_ID, data)

        try:
            res = await self.conn.insert(self.TESTER_SPACE_ID, [1, 'hello2'],
                                         replace=True)

            self.assertListEqual(res.body, [[1, 'hello2']], 'Body ok')
        except Exception as e:
            self.fail(e)

    async def test__insert_invalid_types(self):
        with self.assertRaisesRegex(
                TypeError, r'missing 2 required positional arguments: \'space\' and \'t\''):
            await self.conn.insert()

        with self.assertRaisesRegex(
                TypeError, r'missing 1 required positional argument: \'t\''):
            await self.conn.insert(self.TESTER_SPACE_ID)

    async def test__replace(self):
        data = [1, 'hello']
        res = await self.conn.replace(self.TESTER_SPACE_ID, data)
        self.assertListEqual(res.body, [data], 'Body ok')

        data = [1, 'hello2']
        res = await self.conn.replace(self.TESTER_SPACE_ID, data)
        self.assertListEqual(res.body, [data], 'Body ok')

    async def test__replace_invalid_types(self):
        with self.assertRaisesRegex(
                TypeError, r'missing 2 required positional arguments: \'space\' and \'t\''):
            await self.conn.replace()

        with self.assertRaisesRegex(
                TypeError, r'missing 1 required positional argument: \'t\''):
            await self.conn.replace(self.TESTER_SPACE_ID)



cimport cpython.unicode
from cpython.ref cimport PyObject


cdef bytes encode_unicode_string(str s, bytes encoding=b'utf-8'):
    cdef:
        bytes b
        PyObject *p

    b = <bytes><object>cpython.unicode.PyUnicode_AsEncodedString(
        s, encoding, b'strict'
    )
    return b



import asyncio
import enum
import functools
import os

from asynctnt.exceptions import TarantoolDatabaseError, \
    ErrorCode, TarantoolNotConnectedError
from asynctnt.iproto import protocol
from asynctnt.log import logger

__all__ = (
    'Connection', 'connect', 'ConnectionState'
)


class ConnectionState(enum.IntEnum):
    CONNECTING = 0
    CONNECTED = 1
    RECONNECTING = 2
    DISCONNECTING = 3
    DISCONNECTED = 4


class Connection:
    __slots__ = (
        '_host', '_port', '_username', '_password',
        '_fetch_schema', '_auto_refetch_schema', '_initial_read_buffer_size',
        '_encoding', '_connect_timeout', '_reconnect_timeout',
        '_request_timeout', '_loop', '_state', '_state_prev',
        '_transport', '_protocol', '_db',
        '_disconnect_waiter', '_reconnect_coro'
    )

    def __init__(self, *,
                 host='127.0.0.1',
                 port=3301,
                 username=None,
                 password=None,
                 fetch_schema=True,
                 auto_refetch_schema=True,
                 connect_timeout=60,
                 request_timeout=-1,
                 reconnect_timeout=1. / 3.,
                 encoding=None,
                 initial_read_buffer_size=None,
                 loop=None):

        """
            Connection constructor.
        :param host:
                Tarantool host (pass 'unix/' to connect to unix socket)
        :param port:
                Tarantool port
                (pass '/path/to/sockfile' to connect ot unix socket)
        :param username:
                Username to use for auth
                (if None you are connected as a guest)
        :param password:
                Password to use for auth
        :param fetch_schema:
                Pass True to be able to use spaces and indexes names in
                data manipulation routines (default is True)
        :param auto_refetch_schema:
                If set to True then when ER_WRONG_SCHEMA_VERSION error occurs
                on a request, schema is refetched and the initial request
                is resent. If set to False then schema will not be checked by
                Tarantool, so no errors will occur
        :param connect_timeout:
                Time in seconds how long to wait for connecting to socket
        :param request_timeout:
                Request timeout (in seconds) for all requests
                (by default there is no timeout)
        :param reconnect_timeout:
                Time in seconds to wait before automatic reconnect
                (set to 0 or None to disable auto reconnect)
        :param encoding:
                The encoding to use for all strings
                encoding and decoding (default is 'utf-8')
        :param initial_read_buffer_size:
                Initial and minimum size of read buffer in bytes.
                Higher value means less reallocations, but higher
                memory usage. Lower values
        :param loop:
                Asyncio event loop to use
        """
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._fetch_schema = False if fetch_schema is None else fetch_schema
        if auto_refetch_schema:  # None hack
            self._auto_refetch_schema = True
            if not self._fetch_schema:
                logger.warning('Setting fetch_schema to True as '
                               'auto_refetch_schema is True')
                self._fetch_schema = True
        else:
            self._auto_refetch_schema = False
        self._initial_read_buffer_size = initial_read_buffer_size
        self._encoding = encoding or 'utf-8'

        self._connect_timeout = connect_timeout
        self._reconnect_timeout = reconnect_timeout or 0
        self._request_timeout = request_timeout

        self._loop = loop or asyncio.get_event_loop()

        self._transport = None
        self._protocol = None
        self._db = DbMock()

        self._state = ConnectionState.DISCONNECTED
        self._state_prev = ConnectionState.DISCONNECTED
        self._disconnect_waiter = None
        self._reconnect_coro = None

    def _set_state(self, new_state):
        if self._state != new_state:
            logger.debug('Changing state %s -> %s',
                         self._state.name, new_state.name)
            self._state_prev = self._state
            self._state = new_state
            return True
        return False

    def connection_made(self):
        self._set_state(ConnectionState.CONNECTED)

    def connection_lost(self, exc):
        if self._transport:
            self._transport.close()
        self._transport = None

        if self._reconnect_timeout > 0 \
                and self._state != ConnectionState.DISCONNECTING:
            self._set_state(ConnectionState.DISCONNECTING)
            self._set_state(ConnectionState.RECONNECTING)
            logger.info('%s Started reconnecting', self.fingerprint)
            self._reconnect_coro = \
                asyncio.ensure_future(self._connect(return_exceptions=False),
                                      loop=self._loop)
        else:
            self._set_state(ConnectionState.DISCONNECTED)
            if self._disconnect_waiter:
                self._disconnect_waiter.set_result(True)
                self._disconnect_waiter = None

    def protocol_factory(self, connected_fut, cls=protocol.Protocol):
        return cls(host=self._host,
                   port=self._port,
                   username=self._username,
                   password=self._password,
                   fetch_schema=self._fetch_schema,
                   auto_refetch_schema=self._auto_refetch_schema,
                   request_timeout=self._request_timeout,
                   initial_read_buffer_size=self._initial_read_buffer_size,
                   encoding=self._encoding,
                   connected_fut=connected_fut,
                   on_connection_made=self.connection_made,
                   on_connection_lost=self.connection_lost,
                   loop=self._loop)

    async def _connect(self, return_exceptions=True):
        while True:
            try:
                ignore_states = {
                    ConnectionState.CONNECTING,
                    ConnectionState.CONNECTED,
                    ConnectionState.DISCONNECTING,
                }
                if self._state in ignore_states:
                    return

                self._set_state(ConnectionState.CONNECTING)

                connected_fut = _create_future(self._loop)

                if self._host.startswith('unix/'):
                    unix_path = self._port
                    assert unix_path, \
                        'No unix file path specified'
                    assert os.path.exists(unix_path), \
                        'Unix socket `{}` not found'.format(unix_path)

                    conn = self._loop.create_unix_connection(
                        functools.partial(self.protocol_factory,
                                          connected_fut),
                        unix_path
                    )
                else:
                    conn = self._loop.create_connection(
                        functools.partial(self.protocol_factory,
                                          connected_fut),
                        self._host, self._port)

                try:
                    tr, pr = await asyncio.wait_for(
                        conn, timeout=self._connect_timeout, loop=self._loop)
                except (OSError, asyncio.TimeoutError):
                    raise

                try:
                    await connected_fut
                except:
                    tr.close()
                    # ignoring reconnect by on_connection_lost
                    self._set_state(ConnectionState.DISCONNECTING)
                    raise

                logger.info('%s Connected successfully', self.fingerprint)
                self._set_state(ConnectionState.CONNECTED)

                self._transport = tr
                self._protocol = pr
                self._db = self._protocol.get_common_db()
                self._reconnect_coro = None
                self._normalize_api()
                return
            except TarantoolDatabaseError as e:
                if e.code in {ErrorCode.ER_LOADING}:
                    # If Tarantool is still loading then reconnect
                    if self._reconnect_timeout > 0:
                        await self._do_reconnect(e)
                        continue
                if return_exceptions:
                    self._reconnect_coro = None
                    raise e
                else:
                    logger.exception(e)
                    if self._reconnect_timeout > 0:
                        await self._do_reconnect(e)
                        continue
            except Exception as e:
                if self._reconnect_timeout > 0:
                    await self._do_reconnect(e)
                    continue
                if return_exceptions:
                    self._reconnect_coro = None
                    raise e
                else:
                    logger.exception(e)

    async def _do_reconnect(self, exc=None):
        self._set_state(ConnectionState.RECONNECTING)
        logger.warning('Connect to %s failed: %s. Retrying in %f seconds',
                       self.fingerprint,
                       repr(exc) if exc else '',
                       self._reconnect_timeout)

        await asyncio.sleep(self._reconnect_timeout,
                            loop=self._loop)

    def connect(self):
        """
            Connect coroutine
        """
        return self._connect(return_exceptions=True)

    async def disconnect(self):
        """
            Disconnect coroutine
        """
        if self._state in \
                {ConnectionState.DISCONNECTING, ConnectionState.DISCONNECTED}:
            return
        self._set_state(ConnectionState.DISCONNECTING)

        logger.info('%s Disconnecting...', self.fingerprint)
        waiter = _create_future(self._loop)
        if self._reconnect_coro:
            self._reconnect_coro.cancel()
            self._reconnect_coro = None

        if self._transport:
            self._disconnect_waiter = waiter
            self._transport.close()
            self._transport = None
            self._protocol = None
            self._db = DbMock()
        else:
            waiter.set_result(True)
            self._set_state(ConnectionState.DISCONNECTED)
        return await waiter

    def close(self):
        """
            Same as disconnect, but not a coroutine, i.e. it does not wait
            for disconnect to finish.
        """
        if self._state in \
                {ConnectionState.DISCONNECTING, ConnectionState.DISCONNECTED}:
            return
        self._set_state(ConnectionState.DISCONNECTING)
        logger.info('%s Disconnecting...', self.fingerprint)

        if self._reconnect_coro:
            self._reconnect_coro.cancel()
            self._reconnect_coro = None

        if self._transport:
            self._disconnect_waiter = None
            self._transport.close()
            self._transport = None
            self._protocol = None
            self._db = DbMock()
        else:
            self._set_state(ConnectionState.DISCONNECTED)

    async def reconnect(self):
        """
            Reconnect coroutine.
            Just calls disconnect() and connect()
        """
        await self.disconnect()
        await self.connect()

    @property
    def fingerprint(self):
        return 'Tarantool[{}:{}]'.format(self._host, self._port)

    @property
    def host(self):
        """
            Tarantool host
        """
        return self._host

    @property
    def port(self):
        """
            Tarantool port
        """
        return self._port

    @property
    def username(self):
        """
            Tarantool username
        """
        return self._username

    @property
    def password(self):
        """
            Tarantool password
        """
        return self._password

    @property
    def fetch_schema(self):
        """
            fetch_schema flag
        """
        return self._fetch_schema

    @property
    def auto_refetch_schema(self):
        """
            auto_refetch_schema flag
        """
        return self._auto_refetch_schema

    @property
    def encoding(self):
        """
            Connection encoding
        """
        return self._encoding

    @property
    def reconnect_timeout(self):
        """
            Reconnect timeout value
        """
        return self._reconnect_timeout

    @property
    def connect_timeout(self):
        """
            Connect timeout value
        """
        return self._connect_timeout

    @property
    def request_timeout(self):
        """
            Request timeout value
        """
        return self._request_timeout

    @property
    def version(self):
        """
            Protocol version tuple. ex.: (1, 6, 7)
        """
        if self._protocol is None:
            return None
        return self._protocol.get_version()

    @property
    def loop(self):
        """
            Asyncio event loop
        """
        return self._loop

    @property
    def state(self):
        """
            Current connection state
            :rtype: ConnectionState
        """
        return self._state

    @property
    def is_connected(self):
        """
            Check if connection is active
        """
        if self._protocol is None:
            return False
        return self._protocol.is_connected()

    @property
    def schema(self):
        """
            asynctnt.Schema object
        """
        if self._protocol is None:
            return None
        return self._protocol.schema

    @property
    def schema_id(self):
        """
            Tarantool's current schema id
        """
        if self._protocol is None:
            return None
        return self._protocol.schema_id

    @property
    def encoding(self):
        """
            Connection encoding
        """
        return self._encoding

    @property
    def initial_read_buffer_size(self):
        """
            initial_read_buffer_size value
        """
        return self._initial_read_buffer_size

    def refetch_schema(self):
        """
            Coroutine to force refetch schema
        """
        return self._protocol.refetch_schema()

    def ping(self, *, timeout=-1):
        """
            Ping request coroutine
        :param timeout: Request timeout
        """
        return self._db.ping(timeout=timeout)

    def call16(self, func_name, args=None, *, timeout=-1):
        """
            Call16 request coroutine. It is a call with an old behaviour
            (return result of a Tarantool procedure is wrapped into a tuple,
            if needed)
        :param func_name: function name to call
        :param args: arguments to pass to the function (list object)
        :param timeout: Request timeout
        """
        return self._db.call16(func_name, args,
                               timeout=timeout)

    def call(self, func_name, args=None, *, timeout=-1):
        """
            Call request coroutine. It is a call with a new behaviour
            (return result of a Tarantool procedure is not wrapped into
            an extra tuple). If you're connecting to Tarantool with
            version < 1.7, then this call method acts like a call16 method

        :param func_name: function name to call
        :param args: arguments to pass to the function (list object)
        :param timeout: Request timeout
        """
        return self._db.call(func_name, args,
                             timeout=timeout)

    def eval(self, expression, args=None, *, timeout=-1):
        """
            Eval request coroutine.

        :param expression: expression to execute
        :param args: arguments to pass to the function, that will
                     execute your expression (list object)
        :param timeout: Request timeout
        """
        return self._db.eval(expression, args,
                             timeout=timeout)

    def select(self, space, key=None, **kwargs):
        """
            Select request coroutine.

        :param space: space id or space name.
        :param key: key to select
        :param offset: offset to use
        :param limit: limit to use
        :param index: index id or name
        :param iterator:
                one of the following:
                    * iterator id (int number),
                    * asynctnt.Iterator object
                    * string with an iterator name
        :param timeout: Request timeout
        """
        return self._db.select(space, key, **kwargs)

    def insert(self, space, t, *, replace=False, timeout=-1):
        """
            Insert request coroutine.

        :param space: space id or space name.
        :param t: tuple to insert (list object)
        :param replace: performs replace request instead of insert
        :param timeout: Request timeout
        """
        return self._db.insert(space, t,
                               replace=replace, timeout=timeout)

    def replace(self, space, t, *, timeout=-1):
        """
            Replace request coroutine.

        :param space: space id or space name.
        :param t: tuple to insert (list object)
        :param timeout: Request timeout
        """
        return self._db.replace(space, t,
                                timeout=timeout)

    def delete(self, space, key, **kwargs):
        """
            Delete request coroutine.

        :param space: space id or space name.
        :param key: key to delete
        :param index: index id or name
        :param timeout: Request timeout
        """
        return self._db.delete(space, key, **kwargs)

    def update(self, space, key, operations, **kwargs):
        """
            Update request coroutine.

        :param space: space id or space name.
        :param key: key to update
        :param operations:
                Operations list. Please refer to
                https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
        :param index: index id or name
        :param timeout: Request timeout
        """
        return self._db.update(space, key, operations, **kwargs)

    def upsert(self, space, t, operations, **kwargs):
        """
            Update request coroutine.

        :param space: space id or space name.
        :param t: tuple to insert if it's not in space
        :param operations:
                Operations list to use for update if tuple is already in space.
                Please refer to
                https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
        :param timeout: Request timeout
        """
        return self._db.upsert(space, t, operations, **kwargs)

    def _normalize_api(self):
        if (1, 6) <= self.version < (1, 7):
            Connection.call = Connection.call16


def _create_future(loop):
    try:
        return loop.create_future()
    except AttributeError:
        return asyncio.Future(loop=loop)


class DbMock:
    def __getattr__(self, item):
        raise TarantoolNotConnectedError('Tarantool is not connected')


async def connect(**kwargs):
    """
        connect shorthand. See Connect.__doc__ for kwargs details
    :return: asynctnt.Connection object
    """
    c = Connection(**kwargs)
    await c.connect()
    return c



import asyncio
import contextlib
import functools
import inspect
import logging
import os
import time
import unittest

import sys

import asynctnt
from asynctnt.instance import TarantoolInstance, TarantoolDockerInstance

__all__ = (
    'TestCase', 'TarantoolTestCase'
)


@contextlib.contextmanager
def silence_asyncio_long_exec_warning():
    def flt(log_record):
        msg = log_record.getMessage()
        return not msg.startswith('Executing ')

    logger = logging.getLogger('asyncio')
    logger.addFilter(flt)
    try:
        yield
    finally:
        logger.removeFilter(flt)


class TestCaseMeta(type(unittest.TestCase)):

    @staticmethod
    def _iter_methods(bases, ns):
        for base in bases:
            for methname in dir(base):
                if not methname.startswith('test_'):
                    continue

                meth = getattr(base, methname)
                if not inspect.iscoroutinefunction(meth):
                    continue

                yield methname, meth

        for methname, meth in ns.items():
            if not methname.startswith('test_'):
                continue

            if not inspect.iscoroutinefunction(meth):
                continue

            yield methname, meth

    def __new__(mcls, name, bases, ns):
        for methname, meth in mcls._iter_methods(bases, ns):
            @functools.wraps(meth)
            def wrapper(self, *args, __meth__=meth, **kwargs):
                self.loop.run_until_complete(__meth__(self, *args, **kwargs))
            ns[methname] = wrapper

        return super().__new__(mcls, name, bases, ns)


class TestCase(unittest.TestCase, metaclass=TestCaseMeta):
    loop = None

    @classmethod
    def setUpClass(cls):
        if os.environ.get('USE_UVLOOP'):
            import uvloop
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        asyncio.get_child_watcher().attach_loop(loop)
        cls.loop = loop

    @classmethod
    def tearDownClass(cls):
        if cls.loop:
            cls.loop.close()
        asyncio.set_event_loop(None)

    @contextlib.contextmanager
    def assertRunUnder(self, delta):
        st = time.monotonic()
        try:
            yield
        finally:
            if time.monotonic() - st > delta:
                raise AssertionError(
                    'running block took longer than {}'.format(delta))

    @classmethod
    def ensure_future(cls, coro_or_future):
        return asyncio.ensure_future(coro_or_future, loop=cls.loop)

    @classmethod
    def sleep(cls, delay, result=None):
        return asyncio.sleep(delay, result, loop=cls.loop)


class TarantoolTestCase(TestCase):
    DO_CONNECT = True
    LOGGING_LEVEL = logging.WARNING
    LOGGING_STREAM = sys.stderr
    TNT_APP_LUA_PATH = None
    TNT_CLEANUP = True

    tnt = None

    @classmethod
    def read_applua(cls):
        if cls.TNT_APP_LUA_PATH:
            with open(cls.TNT_APP_LUA_PATH, 'r') as f:
                return f.read()

    @classmethod
    def setUpClass(cls):
        TestCase.setUpClass()
        logging.basicConfig(level=cls.LOGGING_LEVEL,
                            stream=cls.LOGGING_STREAM)
        tarantool_docker_version = os.getenv('TARANTOOL_DOCKER_VERSION')
        if tarantool_docker_version:
            print('Running tarantool in docker. Version = {}'.format(
                tarantool_docker_version))
            tnt = TarantoolDockerInstance(
                applua=cls.read_applua(),
                version=tarantool_docker_version,
                loop=cls.loop
            )
        else:
            unix_path = os.getenv('TARANTOOL_LISTEN_UNIX_PATH')
            if not unix_path:
                tnt = TarantoolInstance(
                    applua=cls.read_applua(),
                    cleanup=cls.TNT_CLEANUP,
                    loop=cls.loop
                )
            else:
                tnt = TarantoolInstance(
                    host='unix/',
                    port=unix_path,
                    console_host='127.0.0.1',
                    applua=cls.read_applua(),
                    cleanup=cls.TNT_CLEANUP,
                    loop=cls.loop
                )
        cls.loop.run_until_complete(tnt.start())
        cls.tnt = tnt

    @classmethod
    def tearDownClass(cls):
        if cls.tnt:
            cls.loop.run_until_complete(cls.tnt.stop())
        TestCase.tearDownClass()

    def setUp(self):
        super(TarantoolTestCase, self).setUp()
        if self.DO_CONNECT:
            self.loop.run_until_complete(self.tnt_connect())

    def tearDown(self):
        self.loop.run_until_complete(self.tnt_disconnect())
        super(TarantoolTestCase, self).tearDown()

    async def tnt_connect(self, *,
                          username=None, password=None,
                          fetch_schema=True,
                          auto_refetch_schema=False,
                          connect_timeout=None, reconnect_timeout=1/3,
                          request_timeout=None, encoding='utf-8',
                          initial_read_buffer_size=None):
        self.conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username=username,
            password=password,
            fetch_schema=fetch_schema,
            auto_refetch_schema=auto_refetch_schema,
            connect_timeout=connect_timeout,
            reconnect_timeout=reconnect_timeout,
            request_timeout=request_timeout,
            encoding=encoding,
            initial_read_buffer_size=initial_read_buffer_size,
            loop=self.loop)
        await self.conn.connect()
        return self.conn

    async def tnt_disconnect(self):
        if hasattr(self, 'conn') and self.conn is not None:
            await self.conn.disconnect()
            self.conn = None

    async def tnt_reconnect(self, **kwargs):
        await self.tnt_disconnect()
        await self.tnt_connect(**kwargs)



import asyncio
import asyncio.subprocess
import logging
import math
import os
import random
import shutil
import string

import functools
import yaml

__all__ = (
    'TarantoolInstanceProtocol', 'TarantoolInstance'
)


class TarantoolInstanceProtocol(asyncio.SubprocessProtocol):
    def __init__(self, tnt, on_exit):
        super().__init__()
        self._tnt = tnt
        self._on_exit = on_exit
        self._transport = None

    @property
    def logger(self):
        return self._tnt.logger

    @property
    def pid(self):
        return self._transport.get_pid() if self._transport else None

    def connection_made(self, transport):
        self.logger.info('Process started')
        self._transport = transport

    def pipe_data_received(self, fd, data):
        if not data:
            return
        line = data.decode()
        line = line.replace('\r', '')
        lines = line.split('\n')
        for line in lines:
            line = line.replace('\n', '')
            line = line.strip()
            if line:
                self.logger.info('=> %s', line)

    def process_exited(self):
        return_code = self._transport.get_returncode()
        if callable(self._on_exit):
            self._on_exit(return_code)

    @property
    def returncode(self):
        return self._transport.get_returncode()

    async def wait(self):
        """Wait until the process exit and return the process return code.

        This method is a coroutine."""
        return await self._transport._wait()

    def send_signal(self, signal):
        self._transport.send_signal(signal)

    def terminate(self):
        self._transport.terminate()

    def kill(self):
        self._transport.kill()


class TarantoolInstance:
    def __init__(self, *,
                 host='127.0.0.1',
                 port=3301,
                 console_host=None,
                 console_port=3302,
                 replication_source=None,
                 title=None,
                 logger=None,
                 log_level=5,
                 slab_alloc_arena=0.1,
                 wal_mode='none',
                 root=None,
                 specify_work_dir=True,
                 cleanup=True,
                 initlua_template=None,
                 applua='-- app.lua --',
                 timeout=5.,
                 command_to_run='tarantool',
                 command_args=None,
                 loop=None
                 ):
        """

        :param host: The host which Tarantool instance is going
                     to be listening on (default = 127.0.0.1)
        :param port: The port which Tarantool instance is going
                     to be listening on (default = 3301)
        :param console_host: The host which Tarantool console is going
                             to be listening on (to execute admin commands)
                             (default = host)
        :param console_port: The port which Tarantool console is going
                             to be listening on (to execute admin commands)
                             (default = 3302)
        :param replication_source: The replication source string.
                                   If it's None, then replication_source=nil
                                   in box.cfg
        :param title: Tarantool instance title (substitutes
                      into custom_proc_title). (default = "tnt[host:port]")
        :param logger: logger, where all messages are logged to.
                       default logger name = Tarantool[host:port]
        :param log_level: Tarantool's log_level (default = 5)
        :param slab_alloc_arena: Tarantool's slab_alloc_arena (default = 0.1)
        :param wal_mode: Tarantool's wal_mode (default = 'none')
        :param root: Tarantool's work_dir location
        :param specify_work_dir: Specify or not the workdir of Tarantool
        :param cleanup: do cleanup or not
        :param initlua_template: The initial init.lua template
                                 (default can be found in
                                 _create_initlua_template function)
        :param applua: Any extra lua script (a string)
                       (default = '-- app.lua --')
        :param timeout: Timeout in seconds - how much to wait for tarantool
                        to become active
        :param command_to_run: command exe
        :param command_args: command args
        :param loop: loop instance
        """
        self._loop = loop or asyncio.get_event_loop()

        self._host = host
        self._port = port
        self._console_host = console_host or host
        self._console_port = console_port
        self._replication_source = replication_source
        self._title = title or self._generate_title()
        self._logger = logger or logging.getLogger(self.fingerprint)
        self._log_level = log_level
        self._slab_alloc_arena = slab_alloc_arena
        self._wal_mode = wal_mode
        self._root = root or self._generate_root_folder_name()
        self._specify_work_dir = specify_work_dir
        self._cleanup = cleanup

        self._initlua_template = initlua_template or \
            self._create_initlua_template()
        self._applua = applua
        self._command_to_run = command_to_run
        self._command_args = command_args

        self._timeout = timeout

        self._is_running = False
        self._is_stopping = False
        self._transport = None
        self._protocol = None
        self._last_return_code = None
        self._stop_event = asyncio.Event(loop=self._loop)

    def _random_string(self,
                       length, *,
                       source=string.ascii_uppercase +
                       string.ascii_lowercase +
                       string.digits):
        return ''.join(random.choice(source) for _ in range(length))

    def _generate_title(self):
        return 'tnt[{}:{}]'.format(self._host, self._port)

    def _generate_root_folder_name(self):
        cwd = os.getcwd()
        path = None
        while path is None or os.path.isdir(path):
            folder_name = '__tnt__' + \
                          self._random_string(10)
            path = os.path.join(cwd, folder_name)
        return path

    def _create_initlua_template(self):
        return """
            box.cfg{
              listen = "${host}:${port}",
              wal_mode = "${wal_mode}",
              custom_proc_title = "${custom_proc_title}",
              slab_alloc_arena = ${slab_alloc_arena},
              replication_source = ${replication_source},
              work_dir = ${work_dir},
              log_level = ${log_level}
            }
            box.schema.user.grant("guest", "read,write,execute", "universe",
                                  nil, {if_not_exists = true})
            require('console').listen("${console_host}:${console_port}")
            ${applua}
        """

    def _render_initlua(self):
        template = string.Template(self._initlua_template)
        d = {
            'host': self._host,
            'port': self._port,
            'console_host': self._console_host,
            'console_port': self._console_port,
            'wal_mode': self._wal_mode,
            'custom_proc_title': self._title,
            'slab_alloc_arena': self._slab_alloc_arena,
            'replication_source': 'nil' if not self._replication_source else '"{}"'.format(self._replication_source),  # nopep8
            'work_dir': '"' + self._root + '"' if self._specify_work_dir else 'nil',  # nopep8
            'log_level': self._log_level,
            'applua': self._applua if self._applua else ''
        }
        return template.substitute(d)

    def _save_initlua(self, initlua):
        initlua = initlua.replace(' ' * 4, '')
        initlua_path = os.path.join(self._root, 'init.lua')
        with open(initlua_path, 'w') as f:
            f.write(initlua)
        return initlua_path

    @property
    def logger(self):
        return self._logger

    @property
    def fingerprint(self):
        return 'Tarantool[{}:{}]'.format(self._host, self._port)

    def prepare(self):
        self._last_return_code = None
        self._stop_event.clear()
        os.mkdir(self._root)
        initlua = self._render_initlua()
        initlua_path = self._save_initlua(initlua)
        return initlua_path

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def console_port(self):
        return self._console_port

    @property
    def is_running(self):
        return self._is_running

    @property
    def pid(self):
        return self._protocol.pid if self._protocol else None

    def _on_process_exit(self, return_code):
        self._last_return_code = return_code
        if self._is_stopping:
            return
        self._stop_event.set()
        self.cleanup()

    async def wait_stopped(self):
        return await self._stop_event.wait()

    async def command(self, cmd, print_greeting=True):
        reader, writer = await asyncio.open_connection(
            self._console_host, self._console_port, loop=self._loop
        )

        greeting = (await reader.read(128)).decode()
        if print_greeting:
            self._logger.info(greeting)

        try:
            if isinstance(cmd, str):
                cmd = cmd.encode('utf-8')
            writer.write(cmd + b'\n')
            data = (await reader.readuntil(b'...\n')).decode()
            data = yaml.load(data)
            return data
        finally:
            writer.close()

    async def start(self):
        self._logger.info('Starting Tarantool instance (%s)', self._title)
        initlua_path = self.prepare()
        self._logger.info('Launching process')

        factory = functools.partial(
            TarantoolInstanceProtocol, self, self._on_process_exit)
        if not self._command_args:
            args = [initlua_path]
        else:
            args = self._command_args
        self._transport, self._protocol = await self._loop.subprocess_exec(
            factory,
            self._command_to_run, *args,
            stdin=None,
            stderr=asyncio.subprocess.PIPE
        )

        interval = 0.1
        attempts = math.ceil(self._timeout / interval)
        while attempts > 0:
            if self._protocol is None or \
                            self._protocol.returncode is not None:
                raise RuntimeError(
                    '{} exited unexpectedly with exit code {}'.format(
                        self.fingerprint, self._last_return_code)
                )
            try:
                status = await self.command('box.info.status',
                                            print_greeting=False)
                if status:
                    status = status[0]
                    if status == 'running':
                        self._logger.info('Moved to the running state')
                        break
            except OSError:
                pass
            await asyncio.sleep(interval, loop=self._loop)
            attempts -= 1
        else:
            raise asyncio.TimeoutError(
                'Timeout while waiting for Tarantool to move to running state')
        self._is_running = True

    async def stop(self):
        if self._protocol is not None:
            self._is_stopping = True
            self._protocol.terminate()
            await self._stop()

    def terminate(self):
        if self._protocol is not None:
            self._is_stopping = True
            self._protocol.terminate()
            self.cleanup()

    def kill(self):
        if self._protocol is not None:
            self._is_stopping = True
            self._protocol.kill()
            self.cleanup()

    async def _stop(self):
        if not self._is_running:
            return

        self._logger.info('Waiting for process to complete')
        await self._protocol.wait()
        self.cleanup()

    def cleanup(self):
        return_code = self._protocol.returncode
        self._logger.info('Finished with return code %d', return_code)

        self._is_running = False
        self._is_stopping = False
        if self._transport:
            self._transport.close()
        self._transport = None
        self._protocol = None
        self._stop_event.clear()
        if self._cleanup:
            shutil.rmtree(self._root, ignore_errors=True)
        self._logger.info('Destroyed Tarantool instance (%s)', self._title)

    def __del__(self):
        self.terminate()


class TarantoolDockerInstance(TarantoolInstance):
    def __init__(self, *,
                 version='1.7',
                 host='0.0.0.0',
                 port=3301,
                 console_port=3302,
                 replication_source=None,
                 title=None,
                 logger=None,
                 log_level=5,
                 slab_alloc_arena=0.1,
                 wal_mode='none',
                 initlua_template=None,
                 applua='-- app.lua --',
                 timeout=10.,
                 loop=None):
        super().__init__(host=host, port=port, console_port=console_port,
                         replication_source=replication_source,
                         title=title, logger=logger, log_level=log_level,
                         slab_alloc_arena=slab_alloc_arena,
                         wal_mode=wal_mode,
                         root=None, specify_work_dir=False, cleanup=True,
                         initlua_template=initlua_template,
                         applua=applua, timeout=timeout,
                         loop=loop)
        self._docker_tarantool_version = version

        cmd = "docker run --rm " \
              "-p {port}:{port} " \
              "-p {console_port}:{console_port} " \
              "-v {root}:/opt/tarantool " \
              "tarantool/tarantool:{version} " \
              "tarantool /opt/tarantool/init.lua"
        cmd = cmd.format(
            port=self.port,
            console_port=self.console_port,
            root=self._root,
            version=self._docker_tarantool_version
        )
        args = cmd.split(' ')
        self._command_to_run = args[0]
        self._command_args = args[1:]



cimport cpython.unicode

from libc.stdint cimport uint32_t, uint64_t, int64_t

cimport tnt

from asynctnt.log import logger
import yaml


cdef class Response:
    @staticmethod
    cdef inline Response new(bytes encoding):
        cdef Response resp
        resp = Response.__new__(Response)
        resp.sync = 0
        resp.code = 0
        resp.schema_id = -1
        resp.errmsg = None
        resp.body = None
        resp.encoding = encoding
        return resp

    cdef inline has_schema_id(self):
        return self.schema_id != -1

    cdef inline is_error(self):
        return self.code != 0

    def __repr__(self):
        return '<Response: code={}, sync={}>'.format(self.code, self.sync)

    def body2yaml(self):
        return yaml.dump(self.body, allow_unicode=True)


cdef object _decode_obj(const char** p, bytes encoding):
    cdef:
        uint32_t i
        mp_type obj_type

        const char *s
        uint32_t s_len

        uint32_t arr_size
        list arr

        uint32_t map_size
        dict map
        mp_type map_key_type
        const char *map_key_str
        uint32_t map_key_len
        object map_key

    obj_type = mp_typeof(p[0][0])
    if obj_type == MP_UINT:
        return mp_decode_uint(p)
    elif obj_type == MP_INT:
        return mp_decode_int(p)
    elif obj_type == MP_STR:
        s = NULL
        s_len = 0
        s = mp_decode_str(p, &s_len)
        return s[:s_len].decode(encoding)
    elif obj_type == MP_BIN:
        s = NULL
        s_len = 0
        s = mp_decode_bin(p, &s_len)
        return <bytes>s[:s_len]
    elif obj_type == MP_BOOL:
        return mp_decode_bool(p)
    elif obj_type == MP_FLOAT:
        return mp_decode_float(p)
    elif obj_type == MP_DOUBLE:
        return mp_decode_double(p)
    elif obj_type == MP_ARRAY:
        arr_size = mp_decode_array(p)
        value = []
        for i in range(arr_size):
            value.append(_decode_obj(p, encoding))
        return value
    elif obj_type == MP_MAP:
        map = {}
        map_size = mp_decode_map(p)
        for i in range(map_size):
            map_key_type = mp_typeof(p[0][0])
            if map_key_type == MP_STR:
                map_key_len = 0
                map_key_str = mp_decode_str(p, &map_key_len)
                map_key = <object>(map_key_str[:map_key_len].decode(encoding))
            elif map_key_type == MP_UINT:
                map_key = <object>mp_decode_uint(p)
            elif map_key_type == MP_INT:
                map_key = <object>mp_decode_int(p)
            else:
                mp_next(p)  # skip current key
                mp_next(p)  # skip value
                logger.warning('Unexpected key type in map: %s',
                               map_key_type)

            map[map_key] = _decode_obj(p, encoding)

        return map
    elif obj_type == MP_NIL:
        mp_next(p)
        return None
    else:
        mp_next(p)
        logger.warning('Unexpected obj type: %s', obj_type)
        return None


cdef list _response_parse_body_data(const char *b, bytes encoding):
    cdef:
        uint32_t size
        uint32_t tuple_size
        list tuples
        uint32_t i, k

    size = mp_decode_array(&b)
    tuples = []
    for i in range(size):
        tuples.append(_decode_obj(&b, encoding))

    return tuples


cdef Response response_parse(const char *buf, uint32_t buf_len,
                             bytes encoding):
    cdef:
        const char *b
        uint32_t size
        uint32_t key
        uint32_t s_len
        const char *s
        Response resp

    b = <const char*>buf
    if mp_typeof(b[0]) != MP_MAP:
        raise TypeError('Response header must be a MP_MAP')

    resp = Response.new(encoding)

    # parsing header
    size = mp_decode_map(&b)
    for _ in range(size):
        if mp_typeof(b[0]) != MP_UINT:
            raise TypeError('Header key must be a MP_UINT')

        key = mp_decode_uint(&b)
        if key == tnt.TP_CODE:
            if mp_typeof(b[0]) != MP_UINT:
                raise TypeError('code type must be a MP_UINT')

            resp.code = <uint32_t>mp_decode_uint(&b)
            resp.code &= 0x7FFF
        elif key == tnt.TP_SYNC:
            if mp_typeof(b[0]) != MP_UINT:
                raise TypeError('sync type must be a MP_UINT')

            resp.sync = mp_decode_uint(&b)
        elif key == tnt.TP_SCHEMA_ID:
            if mp_typeof(b[0]) != MP_UINT:
                raise TypeError('schema_id type must be a MP_UINT')

            resp.schema_id = mp_decode_uint(&b)
        else:
            logger.warning('Unknown argument in header. Skipping.')
            mp_next(&b)

    # parsing body
    if b == &buf[buf_len]:
        # buffer exceeded
        return resp

    if mp_typeof(b[0]) != MP_MAP:
        raise TypeError('Response body must be a MP_MAP')

    size = mp_decode_map(&b)
    for _ in range(size):
        if mp_typeof(b[0]) != MP_UINT:
            raise TypeError('Header key must be a MP_UINT')

        key = mp_decode_uint(&b)
        if key == tnt.TP_ERROR:
            if mp_typeof(b[0]) != MP_STR:
                raise TypeError('errstr type must be a MP_STR')

            s = NULL
            s_len = 0
            s = mp_decode_str(&b, &s_len)
            resp.errmsg = s[:s_len].decode(encoding)
        elif key == tnt.TP_DATA:
            if mp_typeof(b[0]) != MP_ARRAY:
                raise TypeError('body data type must be a MP_ARRAY')
            resp.body = _response_parse_body_data(b, encoding)

    return resp



cimport cython
from libc.stdint cimport uint64_t, int64_t

cimport tnt


@cython.final
@cython.freelist(_BUFFER_FREELIST_SIZE)
cdef class Request:
    @staticmethod
    cdef inline Request new(tnt.tp_request_type op,
                            uint64_t sync, int64_t schema_id,
                            WriteBuffer buf_body):
        cdef Request req
        req = Request.__new__(Request)
        req.op = op
        req.sync = sync
        req.schema_id = schema_id
        req.buf_body = buf_body
        req.waiter = None
        req.timeout_handle = None
        return req

    cdef WriteBuffer build(self):
        cdef WriteBuffer buf
        buf = WriteBuffer.new()
        buf.write_header(self.sync, self.op, self.schema_id)
        if self.buf_body is not None:
            buf.write_buffer(self.buf_body)
        buf.write_length()
        return buf

    def __repr__(self):
        return '<Request op={}, sync={}, schema_id={}, body_len={}>'.format(
            self.op,
            self.sync,
            self.schema_id,
            self.buf_body.len()
        )
